"""
Copyright 2021 Yutong Sean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from enum import Enum
from thbase.hbase.ttypes import TApplicationException, TIOError, TIllegalArgument
from thrift.transport.TTransport import TTransportException
import abc
import logging
logger = logging.getLogger(__name__)


class Observer(object):
    """
    Abstract class for Observer design pattern.
    """
    def __init__(self, target):
        self.target = target

    @abc.abstractmethod
    def handle(self, **kwargs):
        pass


MessageType = Enum('MessageType', ["ERROR"])


class ExceptionHandler(Observer):
    """
    This class responses to all exceptions that caught in the client operations,
    including network exceptions, thrift exceptions and HBase exceptions.
    """
    def __init__(self, target):
        super(ExceptionHandler, self).__init__(target)

    def handle(self, message_type, value):
        """
        The main method to deal with Exceptions.
        TODO:Support more kinds of Exceptions.
        Args:
            message_type: Defined by the enum MessageType.
            value: The message should be scoped with.

        Returns:True if the exception fixed. If not, the exception will be raised further.

        """
        if message_type not in MessageType:
            raise ValueError("Unknown message type.")
        if message_type == MessageType.ERROR:
            if isinstance(value, TIOError):
                error_str = "Connection errors occurred between thrift server and hbase server." \
                            "The error message is: {}".format(value.message)
                try:
                    if value.canRetry:
                        logger.warning(error_str + " The error may be solved by retrying."
                                                   " Client will resend the request.")
                        return True
                    else:
                        logger.error(error_str + " The error cannot be solved by resend the request, client will shutdown.")
                        raise value
                except AttributeError:
                    raise AttributeError("The IOError does not contain the canRetry mark. The error will be raised. "
                                         "Please check if the client version is too old.")
            if isinstance(value, TTransportException):
                if value.type == TTransportException.NOT_OPEN or value.type == TTransportException.TIMED_OUT:
                    logger.warning("A transport error occurs, the message is: {}".format(value.message))
                    logger.warning("System will try to rebuild the connection to solve it.")
                    self.target.connection._reconnect()
                    self.target._refresh_client()
                    return True
                if value.type == TTransportException.ALREADY_OPEN:
                    logger.error("A transport error occurs. The message is: {}".format(value.message))
                if value.type == TTransportException.INVALID_CLIENT_TYPE:
                    logger.error("A transport error occurs. The message is: {}".format(value.message))
                if value.type == TTransportException.NEGATIVE_SIZE:
                    logger.error("The server read a frame with negative length. " +
                                 "Please check your encoding charset and data format.\n" +
                                 "Detailed information is {}.".format(value.message))
                if value.type == TTransportException.SIZE_LIMIT:
                    logger.error("A transport error occurs. The message is: {}".format(value.message))
                if value.type == TTransportException.END_OF_FILE:
                    logger.error("A transport error occurs. The message is: {}".format(value.message))
                if value.type == TTransportException.UNKNOWN:
                    messages = ["Unknown error occurs with tcp transport. ",
                                "This may occur with the following reasons:\n",
                                "1. There is a transport or protocol mismatch between the server and the client.\n",
                                "2. The queue of the server is full.\n",
                                "3. The volume of the sent requests exceed the volume limit of the server.\n"]
                    logger.error("".join(messages))
                raise value
            elif isinstance(value, TApplicationException):
                if value.type == TApplicationException.INVALID_MESSAGE_TYPE:
                    logger.error("An thrift internal error occurs. The message is: {}".format(value.message))
                if value.type == TApplicationException.BAD_SEQUENCE_ID:
                    logger.error("An thrift internal error occurs. The message is: {}".format(value.message))
                if value.type == TApplicationException.INVALID_PROTOCOL:
                    logger.error("An thrift internal error occurs. The message is: {}".format(value.message))
                if value.type == TApplicationException.INVALID_TRANSFORM:
                    logger.error("An thrift internal error occurs. The message is: {}".format(value.message))
                if value.type == TApplicationException.INTERNAL_ERROR:
                    logger.error("An thrift internal error occurs. The message is: {}".format(value.message))
                if value.type == TApplicationException.UNKNOWN_METHOD:
                    logger.error("{}. \n" +
                                 "This error happens when the client code outdated. Please update the client code "
                                 "first.\n" +
                                 "If the problem still exists, please contact with the responsive person.\n")
                raise value
            elif isinstance(value, TIllegalArgument):
                logger.error(value.message)
                logger.error("This error is usually caused by invalid arguments. "
                             "For example, in the Scan operation the start row does not exist. "
                             "Please check the attributes of your Operation object and try again.")
                raise value




"""
Copyright 2021 Yutong Sean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from thbase.util.handlers import ExceptionHandler, MessageType
from thbase.hbase.ttypes import TException

import logging
import time

logger = logging.getLogger(__name__)


class Executor(object):
    def __init__(self, retry_times, retry_timeout, master):
        self._retry_times = retry_times
        self._retry_timeout = retry_timeout
        self._master = master
        self.handler = ExceptionHandler(self._master)

    def call(self, func):
        """
        Execute a specific function with given parameters.
        If the system meets an exception the exception handler will try to handle it.
        If the exception is critical and can not be handled, it will be raised and interrupt the system.
        If the exception is acceptable, the system will retry the operation for several times (set in config).
        After retrying, if the system still cannot get a valid value,
        the function will be marked as failed and return a False.
        Args:
            func: A callable function.

        Returns:
            For Put, Delete operations: True if success, False otherwise.
            For Get, Scan operations: A list of results, False otherwise.
        """
        if not callable(func):
            raise ValueError("A callable function needed here but got {}.".format(type(func)))
        for i in range(self._retry_times + 1):
            try:
                result = func()
                # if result is None, the func is Put, or Delete,
                # so should return a bool to represent if the operation successes.
                if result is None:
                    return True
                # when result is not None, it contains a list of TResult objects with Get and Scan operation.
                # so here return the list directly.
                return result
            except TException as e:
                if not self.handler.handle(MessageType.ERROR, value=e):
                    logger.error("There occurs an error that can not be handled. {}. "
                                 "System is shutdown.".format(e.message))
                    raise e
                else:
                    logger.warn("An error occurs, {}. Redo the operation after {} seconds.".
                                format(e.message, self._retry_timeout))
            time.sleep(self._retry_timeout)
        # this False means the operation failed after retry N times.
        return False



"""
Copyright 2021 Yutong Sean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from thbase.hbase.ttypes import TTableName
from thbase.util.bytes import to_bytes, to_str

__all__ = ['executor', 'handlers', 'type_check', 'check_none', 'str_to_tablename']

DELIMITER = ':'


def type_check(var, t):
    if not isinstance(var, t):
        raise TypeError("A {} object is needed, but got a {}.".format(t.__class, type(var)))


def check_none(var, ms):
    if var is None:
        raise ValueError(ms)


def str_to_tablename(name):
    check_none(name, "")
    type_check(name, str)
    names = name.split(DELIMITER)
    if len(names) == 1:
        return TTableName(ns=to_bytes('default'), qualifier=to_bytes(names[0]))
    elif len(names) == 2:
        return TTableName(ns=to_bytes(names[0]), qualifier=to_bytes(names[1]))
    else:
        raise RuntimeError("Get table name with wrong format.")


def tablename_to_str(table_name):
    type_check(table_name, TTableName)
    return DELIMITER.join([to_str(table_name.ns), to_str(table_name.qualifier)])



"""
Copyright 2021 Yutong Sean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from thbase.hbase import THBaseService
from thbase.hbase.ttypes import TTableDescriptor, TPermissionScope
from thbase.hbase.ttypes import TColumnFamilyDescriptor
from thbase.hbase.ttypes import TAccessControlEntity
from thbase.clientbase import ClientBase
from thbase.thrift2.table import Table
from thbase.util.executor import Executor
from thbase.util.bytes import to_str
from thbase.util.builder import ColumnDescriptorBuilder
from thbase.util.builder import TableDescriptorBuilder
from thbase.util import type_check
from thbase.util import check_none
from thbase.util import str_to_tablename
import logging

logger = logging.getLogger(__name__)


class Client(ClientBase):
    """
    Client implemented by thrift2 API.
    User should not invoke methods of a Client object directly.
    This class does not provide retry mechanism, reconnection mechanism and exception handling.
    Please not use it directly.
    """
    def __init__(self, conf):
        super(Client, self).__init__(conf=conf)
        self.client = THBaseService.Client(self.connection.protocol)
        self.executor = Executor(self.conf.retry_times, self.conf.retry_timeout, master=self)

    def _put_row(self, **kwargs):
        """
        Private method, should not be used by users.
        Args:
            **kwargs:

        Returns:

        """
        table_name = kwargs['table_name']
        put = kwargs['put']
        return self.executor.call(lambda: self.client.put(table_name, put.core))

    def _put_rows(self, **kwargs):
        """
        Private method, should not be used by users.
        Args:
            **kwargs:

        Returns:

        """
        table_name = kwargs['table_name']
        puts = kwargs['puts']
        return self.executor.call(lambda: self.client.putMultiple(table_name, [put.core for put in puts]))

    def _get_row(self, **kwargs):
        """
        Private method, should not be used by users.
        Args:
            **kwargs:

        Returns:

        """
        table_name = kwargs['table_name']
        get = kwargs['get']
        result = self.executor.call(lambda: self.client.get(table_name, get.core))
        return [result]

    def _get_rows(self, **kwargs):
        """
        Private method, should not be used by users.
        Args:
            **kwargs:

        Returns:

        """
        table_name = kwargs['table_name']
        gets = kwargs['gets']
        return self.executor.call(lambda: self.client.getMultiple(table_name, [get.core for get in gets]))

    def _scan(self, **kwargs):
        """
        Private method, should not be used by users.
        Args:
            **kwargs:

        Returns:

        """
        table_name = kwargs['table_name']
        scan = kwargs['scan']
        return self.executor.call(lambda: self.client.getScannerResults(table_name, scan.core, scan.num_rows))

    def _delete_row(self, **kwargs):
        """
        Private method, should not be used by users.
        Args:
            **kwargs:

        Returns:

        """
        table_name = kwargs['table_name']
        delete = kwargs['delete']
        return self.executor.call(lambda: self.client.deleteSingle(table_name, delete.core))

    def _delete_batch(self, **kwargs):
        """
        Private method, should not be used by users.
        Args:
            **kwargs:

        Returns:

        """
        table_name = kwargs['table_name']
        deletes = kwargs['deletes']
        return self.executor.call(lambda: self.client.deleteMultiple(table_name, [delete.core for delete in deletes]))

    def _refresh_client(self):
        """
        Private method, should not be used by users.
        Args:
            **kwargs:

        Returns:

        """
        self.client = THBaseService.Client(self.connection.protocol)

    def get_table(self, table_name):
        """
        Acquire a table object to use functional methods.
        Args:
            **kwargs:

        Returns:

        """
        return Table(table_name=table_name, client=self)

    def create_table(self, desc, split_keys):
        """

        Args:
            desc: TTableDescriptor, which contains the meta information of the table to create.
            split_keys: split keys for table pre-split.

        Returns: True if success, else False.

        """
        type_check(desc, TTableDescriptor)
        type_check(split_keys, list)
        return self.executor.call(lambda: self.client.createTable(desc, split_keys))

    def delete_table(self, table_name):
        """

        Args:
            table_name: The name of the table that need to be removed.

        Returns: True if success, else False.

        """
        tn = str_to_tablename(table_name)
        return self.executor.call(lambda: self.client.deleteTable(tn))

    def enable_table(self, table_name):
        """

        Args:
            table_name: The name of the table that need to be enabled.
            If the table is already enabled, it will raise an Error.
        Returns: True if success, else False.

        """
        tn = str_to_tablename(table_name)
        return self.executor.call(lambda: self.client.enableTable(tn))

    def disable_table(self, table_name):
        """

        Args:
            table_name: The name of the table that need to be disabled.
            If the table is already enabled, it will raise an Error.
        Returns: True if success, else False.

        """
        tn = str_to_tablename(table_name)
        return self.executor.call(lambda: self.client.disableTable(tn))

    def truncate_table(self, table_name, preserve_splits):
        """

        Args:
            table_name:
            preserve_splits:

        Returns:

        """
        tn = str_to_tablename(table_name)
        return self.executor.call(lambda: self.client.truncateTable(tn, preserve_splits))

    def is_enabled(self, table_name):
        """
        Check if the table is enabled.
        Args:
            table_name:

        Returns:

        """
        tn = str_to_tablename(table_name)
        return self.executor.call(lambda: self.client.isTableEnabled(tn))

    def get_tableDescriptor(self, table_name):
        """

        Args:
            table_name:

        Returns:

        """
        tn = str_to_tablename(table_name)
        return self.executor.call(lambda: self.client.getTableDescriptor(tn))

    def get_columnDescriptor(self, table_name, cf):
        """
        Return a column descriptor for specific table with the given cf name.
        Args:
            table_name:
            cf:

        Returns:

        """
        type_check(cf, str)
        td = self.get_tableDescriptor(table_name)
        check_none(td, "Unknown table.")
        for col in td.columns:
            if to_str(col.name) == cf:
                return col
        raise RuntimeError("The table does not contain a column family with name {}".format(cf))

    def get_columnBuilder(self, table_name, cf):
        """
        Acquire a column descriptor of a specific cf.
        Args:
            table_name:
            cf:

        Returns:

        """
        desc = self.get_columnDescriptor(table_name, cf)
        builder = ColumnDescriptorBuilder('')
        builder.copy_from_exist(desc)
        return builder

    def modify_columnFamily(self, table_name, desc):
        """

        Args:
            table_name:
            desc:

        Returns:

        """
        type_check(desc, TColumnFamilyDescriptor)
        tn = str_to_tablename(table_name)
        return self.executor.call(lambda: self.client.modifyColumnFamily(tn, desc))

    def get_tableBuilder(self, table_name):
        """

        Args:
            table_name:

        Returns:

        """
        builder = TableDescriptorBuilder(table_name)
        desc = self.get_tableDescriptor(table_name)
        builder.copy_from_exist(desc)
        return builder

    def modify_table(self, table_descriptor):
        """

        Args:
            table_descriptor:

        Returns:

        """
        type_check(table_descriptor, TTableDescriptor)
        return self.executor.call(lambda: self.client.modifyTable(table_descriptor))

    def grant(self, info):
        """

        Args:
            info:

        Returns:

        """
        type_check(info, TAccessControlEntity)
        return self.executor.call(lambda: self.client.grant(info))

    def revoke(self, info):
        """

        Args:
            info:

        Returns:

        """
        type_check(info, TAccessControlEntity)
        return self.executor.call(lambda: self.client.revoke(info))

    def get_user_permissions(self, domain_name):
        """

        Args:
            domain_name:
            scope:

        Returns:

        """
        type_check(domain_name, str)
        scope = TPermissionScope.TABLE if domain_name[0] != '@' else TPermissionScope.NAMESPACE
        return self.executor.call(lambda: self.client.getUserPermission(domain_name, scope))

    def table_exists(self, table_name):
        """

        Args:
            table_name:

        Returns:

        """
        type_check(table_name, str)
        tn = str_to_tablename(table_name)
        return self.executor.call(lambda: self.client.tableExists(tn))



"""
Copyright 2021 Yutong Sean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from thbase.hbase.ttypes import TTableDescriptor
from thbase.hbase.ttypes import TColumnFamilyDescriptor
from thbase.hbase.ttypes import TCompressionAlgorithm
from thbase.hbase.ttypes import TBloomFilterType
from thbase.hbase.ttypes import TDataBlockEncoding
from thbase.hbase.ttypes import TDurability
from thbase.util.bytes import to_bytes
from thbase.util.bytes import to_str
from thbase.util import type_check
from thbase.util import check_none
from thbase.util import str_to_tablename

COMPRESSION_TYPES = [
    TCompressionAlgorithm.NONE,
    TCompressionAlgorithm.GZ,
    TCompressionAlgorithm.LZ4,
    TCompressionAlgorithm.LZO,
    TCompressionAlgorithm.BZIP2,
    TCompressionAlgorithm.SNAPPY,
    TCompressionAlgorithm.ZSTD,
]

BLOOMFILTER_TYPES = [
    TBloomFilterType.NONE,
    TBloomFilterType.ROW,
    TBloomFilterType.ROWCOL,
    TBloomFilterType.ROWPREFIX_FIXED_LENGTH
]

ENCODING_TYPES = [
    TDataBlockEncoding.NONE,
    TDataBlockEncoding.DIFF,
    TDataBlockEncoding.FAST_DIFF,
    TDataBlockEncoding.PREFIX,
    TDataBlockEncoding.ROW_INDEX_V1
]


class ColumnDescriptorBuilder(object):

    def __init__(self, name=None):
        type_check(name, str)
        self.name = name
        self.attributes = {}
        self.configuration = {}
        self.blockSize = None
        self.bloomnFilterType = None
        self.compressionType = None
        self.dfsReplication = None
        self.dataBlockEncoding = None
        self.keepDeletedCells = None
        self.maxVersions = None
        self.minVersions = None
        self.scope = None
        self.timeToLive = None
        self.blockCacheEnabled = None
        self.cacheBloomsOnWrite = None
        self.cacheDataOnWrite = None
        self.cacheIndexesOnWrite = None
        self.compressTags = None
        self.evictBlocksOnClose = None
        self.inMemory = None

    def build(self):
        return TColumnFamilyDescriptor(name=to_bytes(self.name), attributes=self.attributes,
                                       configuration=self.configuration,
                                       blockSize=self.blockSize, bloomnFilterType=self.bloomnFilterType,
                                       compressionType=self.compressionType, dfsReplication=self.dfsReplication,
                                       dataBlockEncoding=self.dataBlockEncoding, keepDeletedCells=self.keepDeletedCells,
                                       maxVersions=self.maxVersions, minVersions=self.minVersions, scope=self.scope,
                                       timeToLive=self.timeToLive, blockCacheEnabled=self.blockCacheEnabled,
                                       cacheBloomsOnWrite=self.cacheBloomsOnWrite,
                                       cacheDataOnWrite=self.cacheDataOnWrite,
                                       cacheIndexesOnWrite=self.cacheIndexesOnWrite, compressTags=self.compressTags,
                                       evictBlocksOnClose=self.evictBlocksOnClose, inMemory=self.inMemory)

    def copy_from_exist(self, desc):
        type_check(desc, TColumnFamilyDescriptor)
        self.name = desc.name
        self.attributes = desc.attributes
        self.configuration = desc.configuration
        self.blockSize = desc.blockSize
        self.bloomnFilterType = desc.bloomnFilterType
        self.compressionType = desc.compressionType
        self.dfsReplication = desc.dfsReplication
        self.dataBlockEncoding = desc.dataBlockEncoding
        self.keepDeletedCells = desc.keepDeletedCells
        self.maxVersions = desc.maxVersions
        self.minVersions = desc.minVersions
        self.scope = desc.scope
        self.timeToLive = desc.timeToLive
        self.blockCacheEnabled = desc.blockCacheEnabled if desc.blockCacheEnabled else None
        self.cacheBloomsOnWrite = desc.cacheBloomsOnWrite if desc.cacheBloomsOnWrite else None
        self.cacheDataOnWrite = desc.cacheDataOnWrite if desc.cacheDataOnWrite else None
        self.cacheIndexesOnWrite = desc.cacheIndexesOnWrite if desc.cacheIndexesOnWrite else None
        self.compressTags = desc.compressTags if desc.compressTags else None
        self.evictBlocksOnClose = desc.evictBlocksOnClose
        self.inMemory = desc.inMemory if desc.inMemory else None

    def add_attribute(self, key, value):
        check_none(key, "Attribute name is None.")
        if value is None and key in self.attributes:
            self.attributes.pop(key)
        else:
            self.attributes[to_bytes(key)] = to_bytes(value)
        return self

    def add_configuration(self, key, value):
        check_none(key, "Configuration key is None.")
        if value is None and key in self.configuration:
            self.configuration.pop(key)
        else:
            self.configuration[to_bytes(key)] = to_bytes(value)
        return self

    def set_compressionType(self, compress_type):
        if compress_type not in COMPRESSION_TYPES:
            raise RuntimeError("Unknown compression algorithm {}".format(compress_type))
        self.compressionType = compress_type
        return self

    def set_blockSize(self, bs):
        type_check(bs, int)
        if bs < 1:
            raise RuntimeError("Illegal block size " + bs)
        self.blockSize = bs
        return self

    def set_bloomnFilterType(self, bf):
        if bf not in BLOOMFILTER_TYPES:
            raise RuntimeError("Unknown bloom filter type {}".format(bf))
        self.bloomnFilterType = bf
        return self

    def set_dfsReplication(self, hr):
        type_check(hr, int)
        if hr < 1:
            raise RuntimeError("Illegal number of dfs replication {}".format(hr))
        self.dfsReplication = hr

    def set_dataBlockEncoding(self, encoding_type):
        if encoding_type not in ENCODING_TYPES:
            raise RuntimeError("Illegal data block encoding type {}".format(encoding_type))
        self.dataBlockEncoding = encoding_type
        return self

    def set_keepDeletedCells(self, kdc):
        type_check(kdc, bool)
        self.keepDeletedCells = kdc
        return self

    def set_maxVersions(self, mv):
        type_check(mv, int)
        if mv < 1:
            raise RuntimeError("Illegal max versions number {}".format(mv))
        self.maxVersions = mv
        return self

    def set_minVersions(self, mv):
        type_check(mv, int)
        if mv < 0:
            raise RuntimeError("Illegal min versions number {}".format(mv))
        self.minVersions = mv
        return self

    def set_scope(self, scope):
        type_check(scope, int)
        if scope < 0:
            raise RuntimeError("Illegal replication scope value {}".format(scope))
        self.scope = scope
        return self

    def set_timeToLive(self, ttl):
        type_check(ttl, int)
        if ttl < 1:
            raise RuntimeError("Illegal TTL value {}".format(ttl))
        self.timeToLive = ttl
        return self

    def set_blockCacheEnabled(self, enabled):
        type_check(enabled, bool)
        self.blockCacheEnabled = enabled
        return self

    def set_cacheBloomsOnWrite(self, value):
        type_check(value, bool)
        self.cacheBloomsOnWrite = value
        return self

    def set_cacheDataOnWrite(self, value):
        type_check(value, bool)
        self.cacheBloomsOnWrite = value
        return self

    def set_cacheIndexesOnWrite(self, value):
        type_check(value, bool)
        self.cacheIndexesOnWrite = value
        return self

    def set_compressTags(self, value):
        type_check(value, bool)
        self.compressTags = value
        return self

    def set_evictBlocksOnClose(self, value):
        type_check(value, bool)
        self.evictBlocksOnClose = value
        return self

    def set_inMemory(self, value):
        type_check(value, bool)
        self.inMemory = value
        return self


DURABILITY_TYPES = [
    TDurability.SKIP_WAL,
    TDurability.SYNC_WAL,
    TDurability.ASYNC_WAL,
    TDurability.FSYNC_WAL,
    TDurability.USE_DEFAULT
]


class TableDescriptorBuilder(object):

    def __init__(self, name):
        self.name = name
        self.columns = []
        self.attributes = {}
        self.durability = None

    def build(self):
        return TTableDescriptor(tableName=str_to_tablename(self.name),
                                columns=self.columns,
                                attributes=self.attributes,
                                durability=self.durability)

    def copy_from_exist(self, desc):
        type_check(desc, TTableDescriptor)
        self.name = ':'.join([to_str(desc.tableName.ns), to_str(desc.tableName.qualifier)])
        self.columns = desc.columns

    def add_column(self, col):
        type_check(col, TColumnFamilyDescriptor)
        for i in range(len(self.columns)):
            if col.name == self.columns[i].name:
                self.columns[i] = col
                return self
        self.columns.append(col)
        return self

    def add_attributes(self, key, value):
        check_none(key, "None key found.")
        if value is None:
            if key in self.attributes:
                self.attributes.pop(key)
        else:
            self.attributes[to_bytes(key)] = to_bytes(value)
        return self

    def set_durability(self, value):
        if value not in DURABILITY_TYPES:
            raise RuntimeError("Illegal durability type {}".format(value))
        self.durability = value
        return self



"""
Copyright 2021 Yutong Sean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from thbase.config import ClientConfig
from thbase.connection import Connection
from thbase.util.handlers import ExceptionHandler, MessageType
import abc
import logging

logger = logging.getLogger(__name__)


class ClientBase(object):
    """
    Abstract class for both thrift1 and thrift2 client. Implemented with Observer design pattern.
    This class uses a connection object to manage the basic thrift connection with thrift server.
    """
    def __init__(self, conf):
        """
        Initialize the thrift connection and add a new exception handler to deal with exceptions.
        This class should be used by user in NO circumstance!
        Args:
            conf: a customized ClientConfig object.
        """
        if not isinstance(conf, ClientConfig):
            err_str = "Invalid Client Configuration type {}.".format(type(conf))
            logger.error(ValueError(err_str))
            raise ValueError(err_str)
        self.conf = conf
        self.connection = Connection(host=self.conf.host,
                                     port=self.conf.port,
                                     transport_type=self.conf.transport_type,
                                     protocol_type=self.conf.protocol_type,
                                     retry_timeout=self.conf.retry_timeout,
                                     retry_times=self.conf.retry_times,
                                     use_ssl=self.conf.use_ssl,
                                     use_http=self.conf.use_http,
                                     authentication=conf.authentication,
                                     )
        self._observers = set()
        self.attach(ExceptionHandler(self))

    def attach(self, observer):
        """
        Add a new observer into the observer set.
        Args:
            observer: object watching on this client.

        Returns:

        """
        self._observers.add(observer)

    def detach(self, observer):
        """
        Remove an observer from the observer set. If the observer is not registered, do nothing.
        Args:
            observer: object in the observer set.

        Returns:

        """
        if observer in self._observers:
            self._observers.discard(observer)

    def notify(self, message_type, value):
        """
        Notify all the observer to handle something.
        Args:
            message_type: an enum defined in util.handlers module.
            value: the data for handling.

        Returns:

        """
        for obr in self._observers:
            obr.handle(message_type, value)

    def open_connection(self):
        """
        This method open the connection with thrift server.
        Raise TTransport exception.
        Returns: True if success, else False.

        """
        try:
            if not self.connection.is_open():
                self.connection.open()
                return self.connection.is_open()
            return True
        except Exception as e:
            self.notify(MessageType.ERROR, e)
            return self.connection.is_open()

    def close_connection(self):
        """
        This method close the current connection. The close() will not raise any exception and will always success.
        Returns: None

        """
        if self.connection.is_open():
            self.connection.close()

    @abc.abstractmethod
    def _put_row(self, table_name, put):
        """
        Send a single put request to thrift server. Only should be invoked by a Table object.
        Args:
            table_name: Should be invoked by a Table object.
            put: a Put object.

        Returns:
            True if the operation successes, else False.

        """
        pass

    @abc.abstractmethod
    def _put_rows(self, table_name, put_list):
        """
        Send a batch of put requests to thrift server. Only should be invoked by a Table object.
        Args:
            table_name: a str representation of Table name, including the namespace part.
            put_list: a list of Put objects.

        Returns:
            True if the operation successes, else False.

        """
        pass

    @abc.abstractmethod
    def _get_row(self, table_name, get):
        """
        Send a single get request to thrift server. Only should be invoked by a Table object.
        Args:
            table_name: a str representation of Table name, including the namespace part.
            get: a Get object.

        Returns:
            A list of Cells if success. An empty list if the operation fails or the target cell does not exists.
        """
        pass

    @abc.abstractmethod
    def _get_rows(self, table_name, get_list):
        """
        Send a batch of get requests to thrift server. Only should be invoked by a Table object.
        Args:
            table_name: a str representation of Table name, including the namespace part.
            get_list: a list of Get objects.

        Returns:
            A list of Cells if success. An empty list if the operation fails or the target cells do not exists.
        """
        pass

    @abc.abstractmethod
    def _scan(self, table_name, scan):
        """
        Send a scan request to thrift server. Only should be invoked by a Table object.
        Args:
            table_name: a str representation of Table name, including the namespace part.
            scan: a Scan object.

        Returns:
            A list of Cells if success. An empty list if the operation fails or the target cells do not exists.
        """
        pass

    @abc.abstractmethod
    def _delete_row(self, table_name, delete):
        """
        
        Send a delete request to thrift server. Only should be invoked by a Table object.
        Args:
            table_name: a str representation of Table name, including the namespace part.
            delete: a Delete object.

        Returns:
            True if successes, else False.
        """
        pass

    @abc.abstractmethod
    def _delete_batch(self, table_name, delete_list):
        """
        Send a batch of delete requests to thrift server. Only should be invoked by a Table object.
        Args:
            table_name: a str representation of Table name, including the namespace part.
            delete_list: a list of Delete objects.

        Returns:
            True if successes, else False.
        """
        pass

    @abc.abstractmethod
    def _refresh_client(self):
        """
        Reconstruct a client, be used when the client reconnects to the thrift server.
        Returns:
            None
        """
        pass

    @abc.abstractmethod
    def get_table(self, table_name):
        """
        Get a Table object of given table name.
        Args:
            table_name: a str representation of Table name, including the namespace part.

        Returns:
            a Table object.
        """
        pass

    @abc.abstractmethod
    def create_table(self, desc, split_keys):
        """
        Create a table.
        Args:
            desc: a TTableDescriptor that contains the table meta data.
            split_keys: keys for pre-splitting

        Returns:
            True if success, else False.
        """
        pass

    @abc.abstractmethod
    def delete_table(self, table_name):
        """
        Delete a table. The table should be disabled first.
        Args:
            table_name: The corresponding table name.

        Returns:

        """
        pass

    @abc.abstractmethod
    def enable_table(self, table_name):
        """
        Enable a table. If the table is already enabled, will return an error.
        Args:
            table_name: The name of corresponding table.

        Returns:

        """
        pass

    @abc.abstractmethod
    def disable_table(self, table_name):
        """
        Disable a table. If the table is already disabled, will return an error.
        Args:
            table_name: The name of corresponding table.

        Returns:

        """
        pass

    @abc.abstractmethod
    def truncate_table(self, table_name, preserve_splits):
        """
        Drop a table and recreate it.
        Args:
            table_name: The name of the table.
            preserve_splits: If the splits need to be preserved.

        Returns:

        """
        pass

    @abc.abstractmethod
    def is_enabled(self, table_name):
        """
        Check if the given table is enabled.
        Args:
            table_name: TTableName

        Returns:

        """
        pass

    @abc.abstractmethod
    def get_tableDescriptor(self, table_name):
        """
        Get a TTableDescriptor of a specific table.
        Args:
            table_name:

        Returns:

        """
        pass

    @abc.abstractmethod
    def modify_columnFamily(self, table_name, desc):
        """
        Modify the attribute of a column family.
        Args:
            table_name:
            desc:

        Returns:

        """
        pass

    @abc.abstractmethod
    def table_exists(self, table_name):
        """

        Args:
            table_name:

        Returns:

        """
        pass



"""
Copyright 2021 Yutong Sean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import List
from thbase.thrift2.operation import Get, Put, Delete, Scan
from thbase.thrift2.cell import Cell
from thbase.hbase.ttypes import TResult
from thbase.util import type_check
from thbase.util.executor import Executor
from thbase.util.bytes import to_bytes
import logging

logger = logging.getLogger(__name__)


class Table(object):
    """
    This is a class for doing operations on a specific table.
    The object is deprecated to be created by users.
    Use Client.get_table(table_name) method to create a table object.
    """
    def __init__(self, table_name, client):
        self._name = to_bytes(table_name)
        self._client = client
        self.conf = client.conf
        self.executor = Executor(self.conf.retry_times, self.conf.retry_timeout, master=self._client)

    @property
    def name(self):
        # type: () -> str
        return self._name

    def put(self, put):
        # type: (Put) -> bool
        """
        Send a single Put operation to the thrift server.
        Args:
            put: A customized Get object.

        Returns: True if successes, False otherwise.

        """
        type_check(put, Put)
        return self.executor.call(lambda: self._client._put_row(table_name=self.name, put=put))

    def put_batch(self, puts):
        # type: (List[Put]) -> bool
        """
        Send multiple Put requests to the server at one time.
        The requests will be sent batch by batch.
        The logger will log the failed position if one batch of requests failed.
        Args:
            puts: A list of Put objects

        Returns: True if successes, False otherwise.

        """
        type_check(puts, list)
        for put in puts:
            type_check(put, Put)
        for i in range(0, len(puts), self.conf.batch_size):
            result = self._client._put_rows(table_name=self.name, puts=puts[i: i + self.conf.batch_size])
            if not result:
                logger.error("An error occurs at index {}, the Put requests after {} (inclusive) failed.".format(i, i))
                return False
        return True

    def get(self, get):
        # type: (Get) -> List[Cell]
        """
        Send a single Get operation to the thrift server.
        Args:
            get: A customized Get object.

        Returns: If success: a list of cells.
                 If success but not matched data: an empty list.
                 If get failed: False.

        """
        type_check(get, Get)
        result = self._client._get_row(table_name=self.name, get=get)
        # if result is False, that means the operation failed after retry N times.
        # if result is [], that means there is no matched cell in hbase.
        if not result:
            return []
        return self._results_format(result)

    def get_batch(self, gets):
        # type: (List[Get]) -> List[Cell]
        """
        Send multiple Get requests to the server at one time.
        The requests will be sent batch by batch.
        The logger will log the failed position if one batch of requests failed.
        Args:
            gets: A list of Get objects.

        Returns: If success: a list of cells.
                 If success but not matched data: an empty list.
                 If get failed: False.
                 If partly success, a part result will be returned.

        """
        type_check(gets, (list, tuple))
        for get in gets:
            type_check(get, Get)
        result_list = []
        for i in range(0, len(gets), self.conf.batch_size):
            result = self._client._get_rows(table_name=self.name, gets=gets[i: i + self.conf.batch_size])
            # if result == False, it shows that the operation failed.
            # The task should stop and return the successful part.
            if result is False:
                return self._results_format(result_list)
            elif len(result) > 0:
                result_list += result
        return self._results_format(result_list)

    def scan(self, scan):
        # type: (Scan) -> List[Cell]
        """
        Send a Scan request to the thrift server.
        Args:
            scan: A single Scan object.

        Returns:If success: a list of cells.
                 If success but not matched data: an empty list.
                 If the start row do not exists, it will raise an IllegalArgument error.
                 If scan failed: False.

        """
        type_check(scan, Scan)
        return self._results_format(self._client._scan(table_name=self.name, scan=scan))

    def delete(self, delete):
        # type: (Delete) -> bool
        """
        Send a Delete request to the thrift server.
        Args:
            delete: a single Delete object

        Returns: True if successes, False otherwise.

        """
        type_check(delete, Delete)
        return self._client._delete_row(table_name=self.name, delete=delete)

    def delete_batch(self, batch):
        # type: (List[Delete]) -> bool
        """
        Send a list of Delete requests to the thrift server.
        The requests will be sent batch by batch.
        The logger will log the failed position if one batch of requests failed.
        Args:
            batch: a list of Delete objects.

        Returns: True if successes, False otherwise.

        """
        type_check(batch, (list, tuple))
        for delete in batch:
            type_check(delete, Delete)
        for i in range(0, len(batch), self.conf.batch_size):
            if not self._client._delete_batch(table_name=self.name,
                                              deletes=batch[i: i + self.conf.batch_size]):
                logger.error("Delete_batch failed at index {}, the delete requests after {} (inclusive) are not sent.".format(i, i))
                return False
        return True

    def _results_format(self, results):
        # type: (List[TResult]) -> List[Cell]
        """
        @Deprecated
        Inner util method. Should not be used by user.
        Transform the thrift result to a series of Cell objects.
        Args:
            results: a list of TResult.

        Returns: an empty list if there is no results or a list of Cell objects.

        """
        result_list = []  # type: List[Cell]
        if not results or len(results) == 0:
            return []
        for result in results:
            for cv in iter(result.columnValues):
                result_list.append(Cell(self._name, result.row, cv.family, cv.qualifier, cv.value, cv.timestamp))
        return result_list
